/*
 * -----------------------------------------------------------------------
 * Drawing stuff
 */

#ifndef	SRC_ALPHA
#define	SRC_ALPHA	0.5f
#endif

#ifndef	ORIGIN_COLOR
#define	ORIGIN_COLOR	0.8f,0.8f,0,SRC_ALPHA
#endif

#ifndef	MINOR_GRID_COLOR
#define	MINOR_GRID_COLOR	0,0.4f,0,SRC_ALPHA	
#endif

#ifndef	MAJOR_GRID_COLOR
#define	MAJOR_GRID_COLOR	0.2f,0.2f,0,SRC_ALPHA	
#endif

#ifndef	MINOR_GRID_LINE_WIDTH
#define	MINOR_GRID_LINE_WIDTH	0.8f
#endif

#ifndef	MAJOR_GRID_LINE_WIDTH
#define	MAJOR_GRID_LINE_WIDTH	1.7f
#endif

#ifndef MINOR_GRID_DIVS
#define MINOR_GRID_DIVS	100
#endif

#ifndef MAJOR_GRID_DIVS
#define MAJOR_GRID_DIVS	10
#endif

static GLfloat alpha = 0, beta = 0, gamm = 0, delt = 0;
static GLfloat deltax = 0, deltay = 0;
static GLfloat fmin_ = 1 * 4.4f, fcenter = 15 * 4.4f, fmax_ = 29 * 4.4f, ext_scale = 1 * 4 * 1.15f;
static vboolean auto_warp = false, draw_minor_grid = true, draw_major_grid = true, draw_tripod = true;

static void
draw_xOy_grid(vec3_t origin,vec_t width,vec_t height,unsigned num_divs)
{
	unsigned i;
	vec3_t p, dp;
	vec3Copy(p,origin);
	p[0] -= (vec_t)0.5 * width;
	p[1] -= (vec_t)0.5 * height;
	vec3Set(dp,width,height,0);
	vec3Scale(dp,dp,(vec_t)1 / num_divs);
	glBegin(GL_LINES);
		/* grid along Oy direction */
		for(i = 0;i <= num_divs;i++)
		{
			glVertex3d(p[0] + i * dp[0],p[1],p[2]);
			glVertex3d(p[0] + i * dp[0],p[1] + height,p[2]);
		}
		/* grid along Ox direction */
		for(i = 0;i <= num_divs;i++)
		{
			glVertex3d(p[0],p[1] + i * dp[1],p[2]);
			glVertex3d(p[0] + width,p[1] + i * dp[1],p[2]);
		}
	glEnd(); /* GL_LINES */
}

static void
draw_2_xOy_grids(vec3_t origin,vec3_t center,vec_t width,vec_t height)
{
	vec_t dw, dh;
	vec3_t neworg;
	if(!draw_minor_grid && !draw_major_grid)
		return;
	dw = width / (MAJOR_GRID_DIVS - 1);
	dh = height / (MAJOR_GRID_DIVS - 1);
	vec3Copy(neworg,center);
	neworg[0] -= (vec_t)fmod(center[0] - origin[0],dw);
	neworg[1] -= (vec_t)fmod(center[1] - origin[1],dh);
	if(draw_minor_grid)
	{
		glLineWidth(MINOR_GRID_LINE_WIDTH);
		glColor4f(MINOR_GRID_COLOR);
		draw_xOy_grid(neworg,width,height,MINOR_GRID_DIVS);
	}
	if(draw_major_grid)
	{
		glLineWidth(MAJOR_GRID_LINE_WIDTH);
		glColor4f(MAJOR_GRID_COLOR);
		draw_xOy_grid(neworg,width,height,MAJOR_GRID_DIVS);
	}
}

static vboolean
draw(vboolean (*draw_main)(void))
{
	vboolean result;
	vec3_t mins, maxs, origin, center;
	vec_t scale, se;

	/* Compute bounding box */
/*
	int i;
	bboxClear(mins,maxs);
	for(i = 0;i < NUM_OBJECTS;i++)
		bboxAddPoint(mins,maxs,obj[i].pos);
*/
	vec3Set(mins,-1,-1,-1);
	vec3Set(maxs,1,1,1);
	
	/* translate to center of all objects */
	vec3Zero(origin);
	vec3Sub(center,maxs,mins); /* extents */
	scale = 4.0f / vec3Abs(center);
	vec3Add(center,mins,maxs);
	vec3Scale(center,center,0.5f);

	glClear(GL_COLOR_BUFFER_BIT);
	glLoadIdentity();

	glRotatef(gamm,0,0,1);

	glTranslatef(0,0,-fcenter);
	se = scale * ext_scale; /* preserve 'scale' */
	glScaled(se,se,se); 

	glTranslatef(deltax,deltay,0);

	glRotatef(alpha,0,1,0);
	glRotatef(beta,1,0,0);
	glRotatef(delt,0,0,1);

	glTranslated(-center[0],-center[1],-center[2]);

	/* Draw scene */
	glPushMatrix();
	result = draw_main();
	glPopMatrix();
	/* Draw coordsys-aligned grid on xOy plane */
	draw_2_xOy_grids(origin,center,10,10);
	/* Origin and axes */
	if(draw_tripod)
	{
		glLineWidth(4.0f);
		glBegin(GL_LINES);
			glColor4f(.9f,.3f,.3f,SRC_ALPHA);
			glVertex3f(0,0,0);
			glVertex3f(1,0,0);
			glColor4f(.3f,.9f,.3f,SRC_ALPHA);
			glVertex3f(0,0,0);
			glVertex3f(0,1,0);
			glColor4f(.3f,.3f,.9f,SRC_ALPHA);
			glVertex3f(0,0,0);
			glVertex3f(0,0,1);
		glEnd();
		glPointSize(6.0f);
		glBegin(GL_POINTS);
			glColor4f(ORIGIN_COLOR);
			glVertex3fv(origin);
		glEnd();
	}
	agSwap();
	return result;
}

/*
 * -----------------------------------------------------------------------
 * Part 6: Camera Control
 */
static vboolean paused = true;
/*
static int autopause = 0;
*/

static void
correct_frustum(void)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-1,1,-1,1,fmin_,fmax_);
	glMatrixMode(GL_MODELVIEW);
}

static vboolean
handle_event_check_break(ui_event_p in,
	vboolean (*next_check_break)(ui_event_p ev))
{
	switch(in->type)
	{
		case ui_key_event:
			if(in->key.press)
			{
				switch(in->key.scan)
				{
					case sc_escape:
						return 1;
					case sc_wheelup:
					case sc_minus:
					case sc_kp_subtract:
						fmin_ *= 0.9f;
						fmax_ *= 0.9f;
						fcenter = (fmin_ + fmax_) * 0.5f;
						ext_scale *= 0.85f;
						correct_frustum();
						break;
					case sc_wheeldown:
					case sc_equal:
					case sc_kp_add:
						fmin_ *= 1.1f;
						fmax_ *= 1.1f;
						ext_scale *= 1.15f;
						fcenter = (fmin_ + fmax_) * 0.5f;
						correct_frustum();
						break;
					case sc_space:
					case sc_kp_begin:
						alpha = beta = deltax = deltay = gamm = delt = 0;
						break;
					case sc_up:
						deltay -= 0.1f;
						break;
					case sc_kp_up:
						beta += 3;
						break;
					case sc_down:
						deltay += 0.1f;
						break;
					case sc_kp_down:
						beta -= 3;
						break;
					case sc_left:
						deltax += 0.1f;
						break;
					case sc_kp_left:
						alpha -= 3;
						break;
					case sc_right:
						deltax -= 0.1f;
						break;
					case sc_kp_right:
						alpha += 3;
						break;
					case sc_kp_home:
						gamm += 3;
						break;
					case sc_kp_prior:
						gamm -= 3;
						break;
					case sc_kp_end:
						delt += 3;
						break;
					case sc_kp_next:
						delt -= 3;
						break;
					case sc_w:
						auto_warp = !auto_warp;
						uiAutoWarp(auto_warp ? GL_TRUE : GL_FALSE);
						break;
	/*
					case sc_a:
						autopause = !autopause;
						break;
	*/
					case sc_p:
						paused = !paused;
						break;
					case sc_i:
						draw_minor_grid = !draw_minor_grid;
						break;
					case sc_o:
						draw_major_grid = !draw_major_grid;
						break;
					case sc_t:
						draw_tripod = !draw_tripod;
						break;
				}
			}
			break;
		case ui_wm_event:
			switch(in->configure.wm_event)
			{
				case ui_wm_close:
					return true;
			}
			break;
		case ui_pointer_motion_event:
			if((in->button & UI_LBUTTON_MASK) || 
				(auto_warp && !(in->button & UI_RBUTTON_MASK)))
			{
				alpha += in->client_dx;
				beta += in->client_dy;
			}
			if(in->button & UI_RBUTTON_MASK)
			{
				deltax += 0.05f * in->client_dx / ext_scale;
				deltay -= 0.05f * in->client_dy / ext_scale;
			}
			break;
		default:
			break;
	}
	return next_check_break ? next_check_break(in) : false;
}

static void
agMainLoop(vboolean (*draw_main)(void),
	vboolean (*next_check_break)(ui_event_p ev))
{
	vboolean not_done = true;
	static unsigned computation_time = 20; /* 50 FPS max */
	static unsigned lastdraw_time = 0;
	ui_event_t in;
	correct_frustum();
	while(not_done)
	{
		while(uiPending())
		{
			uiNextEvent(&in);
			if(handle_event_check_break(&in,next_check_break))
				goto good_bye;
		}
		if(!paused)
		{
			/* compute_progress(); */
			agSleep(10);
		}
		else
			agSleep(computation_time);
		if(agTickCount() - lastdraw_time > computation_time)
		{
			not_done = draw(draw_main);
			lastdraw_time = agTickCount();
		}
		/* print_status(); */
	}
good_bye:;
}

